package edu.berkeley.qfat.bind;
import edu.berkeley.qfat.geom.*;
import javax.media.opengl.*;
import java.util.*;

/**
 *  An equivalence class of geometric objects whose positions are
 *  related by affine transformation matrices and are constrained by
 *  an affine constraint.
 */
class BindingGroup<T extends HasBindingGroup> implements Iterable<T> {

    private T                  master     = null;
    private AffineConstraint   constraint = new AffineConstraint.All();
    private HashMap<T,Matrix>  matrices   = new HashMap<T,Matrix>();

    public BindingGroup(T master) {
        this.master = master;
        matrices.put(master, Matrix.ONE);
    }

    int size() { return matrices.size(); }

    /** merge another binding group with this one */
    void merge(BindingGroup<T> bg, Matrix m, float epsilon) {
        if (bg==this) {
            if (m.equalsModuloEpsilon(Matrix.ONE, epsilon)) return;
            // FIXME: what if points do not fall on the merged constraint-line?
            constraint = constraint.intersect(m.getAffineConstraint(epsilon), epsilon);
            return;
        }

        for(HasBindingGroup hbg : bg.matrices.keySet()) {
            matrices.put((T)hbg, bg.matrices.get(hbg).times(m));
            hbg.bindingGroup = this;
        }

        // FIXME: what if points do not fall on the merged constraint-line?
        AffineConstraint ac = bg.constraint.multiply(getMatrix(master, bg.master));
        constraint = constraint.intersect(ac, epsilon);

        bg.master = null;
        for(HasBindingGroup hbg : bg.matrices.keySet())
            hbg.bindingGroupChanged();
        bg.matrices.clear();
    }

    public Matrix getMatrix(T t) { return matrices.get(t); }

    public Iterator<T> iterator() { return matrices.keySet().iterator(); }

    /** t1 = getMatrix(t1, t2) * t2 */
    public Matrix getMatrix(T t1, T t2) {
        //                    t1 = getMatrix(t1) * master
        // getMatrix(t2)^-1 * t2 =                 master
        //                    t1 = getMatrix(t1) * getMatrix(t2)^-1 * t2
        return getMatrix(t1).times(getMatrix(t2).inverse());
    }

    public AffineConstraint getConstraint(T t) {
        return constraint.multiply(matrices.get(t));
    }

    public void unbind(T trem) {
        if (trem != master) {
            matrices.remove(trem);
            return;
        }
        if (matrices.size()==1) {
            master = null;
            matrices.remove(trem);
            return;
        }
        Iterator<T> it = iterator();
        T newmaster = it.next();
        if (newmaster==trem) newmaster = it.next();
        if (newmaster==trem) throw new Error();
        HashMap<T,Matrix> newmatrices = new HashMap<T,Matrix>();
        for(T t : matrices.keySet()) {
            if (t==trem) continue;
            newmatrices.put(t, getMatrix(t, newmaster));
        }
        master = newmaster;
        matrices = newmatrices;
    }

    public boolean contains(HasBindingGroup t) {
        return matrices.get((T)t) != null;
    }
}
